import { Diff } from "deep-diff";
import { RuleResult, RuleSignature } from "./rules/rules.js";
import { OpenAPIV2 } from "openapi-types";
import { SuppressionRegistry } from "./suppression.js";
export interface DiffClientConfig {
    lhs: string | string[];
    rhs: string | string[];
    args: any;
    rules?: RuleSignature[];
}
export declare class DiffClient {
    #private;
    args: any;
    suppressions?: SuppressionRegistry;
    private rules;
    private lhsParser?;
    private rhsParser?;
    /** Tracks if shortenKeys has been called to avoid re-running the algorithm needlessly. */
    private keysShortened;
    /** The parsed lhs document. */
    lhs?: OpenAPIV2.Document;
    /** The parsed rhs document. */
    rhs?: OpenAPIV2.Document;
    /** The results of the diff operation. Available once processDiff() called. */
    diffResults?: DiffResult;
    /** The output files generated by the diff operation. Available once buildOutput() called. */
    resultFiles?: ResultFiles;
    /** The temporary root folder for TypeSpec-generated Swagger. */
    private tempRhsRoot?;
    /** Creates an instance of the DiffClient class asynchronously. */
    static create(config: DiffClientConfig): Promise<DiffClient>;
    protected constructor(config: DiffClientConfig);
    /**
     * If the rhsRoot is not provided, rhs points to TypeSpec and the lhs is a folder, use the lhs as the basis
     * for the rhsRoot.
     * @param lhs lhs paths
     * @param rhs rhs paths
     * @returns the inferred rhsRoot or undefined if it cannot be inferred
     */
    private checkDefaultRhsRoot;
    /**
     * Parses the documents, expanding them into canonical transformations.
     * Upon completion, the lhs and rhs documents will be available for diffing.
     */
    parse(): void;
    /**
     * Updates the parsed documents to shorten their keys.
     * Upon completion, the lhs and rhs should have shortened keys
     * and be ready for diffing.
     */
    shortenKeys(): void;
    /**
     * Runs the deep-diff algorithm on the documents and sorts the results into
     * three categories: flaggedViolations, assumedViolations, and noViolations.
     */
    processDiff(): void;
    /**
     * Processes all rules against the given diff. If no rule confirms or denies
     * an issue, the diff is treated as a failure. If a rule is flagged as a violation,
     * it can be overridden by a later rule, but as soon as a rule confirms something
     * is not a violation, processing stops.
     * @param data the diff data to evaluate.
     * @returns an allowed DiffRuleResult. Only "ContinueProcessing" is not allowed.
     */
    private processRules;
    /**
     * Constructs the output files based on the diff results.
     */
    buildOutput(): void;
    /** Returns true if the summary indicates that there are violations. */
    hasViolations(summary: ResultSummary, preserveDefinitions: boolean): boolean;
    /** Write results to output files and print summary to console. */
    writeOutput(): void;
}
/** Describes a diff item */
export interface DiffItem {
    ruleResult: RuleResult;
    ruleName?: string;
    message?: string;
    diff: Diff<any, any>;
}
/** Describes a grouping of diff items. */
export interface DiffGroupingResult {
    name?: string;
    count: number;
    items: DiffItem[];
}
/** Describes the sorted results of diffing. */
interface DiffResult {
    flaggedViolations: DiffItem[];
    assumedViolations: DiffItem[];
    noViolations: DiffItem[];
    suppressedViolations: DiffItem[];
}
interface ResultFiles {
    raw: [any, any];
    normal: [any, any];
    inverse: [any, any];
    diff: any;
    diffInverse: any;
}
interface ResultSummary {
    flaggedViolations: number;
    rulesViolated: number | undefined;
    assumedViolations: number;
    assumedRules: number | undefined;
    unresolvedReferences: number;
    unreferencedObjects: number;
    suppressedViolations: number;
}
export {};
//# sourceMappingURL=diff-client.d.ts.map