import { RuleResult as RuleResult } from "./rules.js";
/**
 * Ensures that the LHS is compatible with the RHS.
 * The LSH must be a subset of the RHS.
 */
function reportIncompatibilities(lhs, rhs, prefix) {
    const errors = [];
    if (!lhs || !rhs) {
        return errors;
    }
    // ensure the types are the same
    if (lhs.type !== rhs.type) {
        errors.push(`LHS type '${lhs.type}' is not the same as RHS type '${rhs.type}'`);
    }
    const lhsRequired = lhs.required ?? [];
    const rhsRequired = rhs.required ?? [];
    const lhsKeys = Object.keys(lhs.properties ?? {});
    const rhsKeys = Object.keys(rhs.properties ?? {});
    // ensure that lhs has all the rhs required keys
    for (const key of rhsRequired) {
        if (!lhsKeys.includes(key)) {
            const path = prefix ? `${prefix}.${key}` : key;
            errors.push(`LHS is missing key '${path}' required by RHS`);
        }
    }
    // ensure every property in lhs is in rhs
    for (const key of lhsKeys) {
        if (!rhsKeys.includes(key)) {
            const path = prefix ? `${prefix}.${key}` : key;
            errors.push(`LHS has key '${path}' which is not in RHS`);
        }
    }
    for (const key of lhsKeys) {
        errors.push(...reportIncompatibilities(lhs.properties[key], rhs.properties[key], prefix ? `${prefix}.${key}` : `${key}`));
    }
    return errors;
}
/**
 * Ignores a diff in the minLength property on the api-version parameter only.
 */
export function compareErrorsRule(data, lhs, rhs) {
    if (!data.path)
        return;
    if (data.kind !== "E")
        return;
    const lastPath = data.path[data.path.length - 1];
    if (lastPath !== "$error")
        return;
    const lhsError = lhs?.definitions?.[data.lhs];
    const rhsError = rhs?.definitions?.[data.rhs];
    const errors = reportIncompatibilities(lhsError, rhsError);
    if (errors.length === 0) {
        return RuleResult.NoViolation;
    }
    else {
        return [RuleResult.FlaggedViolation, errors.join("\n")];
    }
}
//# sourceMappingURL=compare-errors.js.map